from os.path import join
from re import compile as re_compile
from typing import List

from supervisely.nn.artifacts.artifacts import BaseTrainArtifacts


class YOLOv5(BaseTrainArtifacts):
    def __init__(self, team_id: int):
        super().__init__(team_id)

        self._app_name = "Train YOLOv5"
        self._slug = "supervisely-ecosystem/yolov5/supervisely/train"
        self._serve_app_name = "Serve YOLOv5"
        self._serve_slug = "supervisely-ecosystem/yolov5/supervisely/serve"
        self._framework_name = "YOLOv5"
        self._framework_folder = "/yolov5_train"
        self._weights_folder = "weights"
        self._task_type = "object detection"
        self._weights_ext = ".pt"
        self._config_file = None
        self._pattern = re_compile(r"^/yolov5_train/[^/]+/\d+/?$")
        self._available_task_types: List[str] = ["object detection"]
        self._require_runtime = False

    def get_task_id(self, artifacts_folder: str) -> str:
        return artifacts_folder.split("/")[-1]

    def get_project_name(self, artifacts_folder: str) -> str:
        return artifacts_folder.split("/")[-2]

    def get_task_type(self, artifacts_folder: str) -> str:
        return self._task_type

    def get_weights_path(self, artifacts_folder: str) -> str:
        return join(artifacts_folder, self._weights_folder)

    def get_config_path(self, artifacts_folder: str) -> str:
        return None


class YOLOv5v2(YOLOv5):
    def __init__(self, team_id: int):
        super().__init__(team_id)

        self._app_name = "Train YOLOv5 2.0"
        self._slug = "supervisely-ecosystem/yolov5_2.0/train"
        self._serve_app_name = "Serve YOLOv5 2.0"
        self._serve_slug = "supervisely-ecosystem/yolov5_2.0/serve"
        self._framework_name = "YOLOv5 2.0"
        self._framework_folder = "/yolov5_2.0_train"
        self._weights_folder = "weights"
        self._task_type = "object detection"
        self._weights_ext = ".pt"
        self._config_file = None
        self._pattern = re_compile(r"^/yolov5_2.0_train/[^/]+/\d+/?$")
        self._available_task_types: List[str] = ["object detection"]



from __future__ import annotations

from typing import Dict, Literal

import numpy as np
import pandas as pd

from supervisely.nn.benchmark.object_detection.base_vis_metric import DetectionVisMetric
from supervisely.nn.benchmark.visualization.widgets import (
    ChartWidget,
    CollapseWidget,
    ContainerWidget,
    MarkdownWidget,
    RadioGroupWidget,
)


class PerClassOutcomeCounts(DetectionVisMetric):
    MARKDOWN = "per_class_outcome_counts"
    MARKDOWN_2 = "per_class_outcome_counts_2"
    CHART = "per_class_outcome_counts"
    COLLAPSE_TIP = "per_class_outcome_counts_collapse"
    RADIO_GROUP = "per_class_outcome_counts_radio_group"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.clickable = True
        self._keypair_sep: str = "-"
        self.switchable = True

    @property
    def md(self) -> MarkdownWidget:
        text = self.vis_texts.markdown_class_outcome_counts_1
        text = text.format(
            self.vis_texts.definitions.true_positives,
            self.vis_texts.definitions.false_positives,
            self.vis_texts.definitions.false_negatives,
        )
        return MarkdownWidget(self.MARKDOWN, "Outcome Counts by Class", text)

    @property
    def md_2(self) -> MarkdownWidget:
        text = self.vis_texts.markdown_class_outcome_counts_2
        text = text.format(self.vis_texts.definitions.f1_score)
        return MarkdownWidget(self.MARKDOWN_2, "Outcome Counts by Class", text)

    @property
    def collapse(self) -> CollapseWidget:
        md = MarkdownWidget(
            name=self.COLLAPSE_TIP,
            title="Normalization",
            text=self.vis_texts.markdown_normalization,
        )
        return CollapseWidget([md])

    @property
    def chart(self) -> ContainerWidget:
        return ContainerWidget(
            [self.radio_group(), self._get_chart("normalized"), self._get_chart("absolute")],
            self.CHART,
        )

    def radio_group(self) -> RadioGroupWidget:
        return RadioGroupWidget(
            "Normalization",
            self.RADIO_GROUP,
            ["normalized", "absolute"],
        )

    def _get_chart(self, switch_key: Literal["normalized", "absolute"]) -> ChartWidget:
        chart = ChartWidget(
            self.CHART,
            self._get_figure(switch_key),
            switch_key=switch_key,
            switchable=self.switchable,
            radiogroup_id=self.RADIO_GROUP,
        )
        chart.set_click_data(
            self.explore_modal_table.id,
            self.get_click_data(),
            chart_click_extra="'getKey': (payload) => `${payload.points[0].label}${'-'}${payload.points[0].data.name}`, 'keySeparator': '-',",
        )
        return chart

    def _get_figure(self, switch_key: Literal["normalized", "absolute"]):  # -> go.Figure:
        import plotly.express as px  # pylint: disable=import-error

        # Per-class Counts
        mp = self.eval_result.mp
        tp = mp.m._take_iou_thresholds(mp.true_positives).flatten()
        fp = mp.m._take_iou_thresholds(mp.false_positives).flatten()
        fn = mp.m._take_iou_thresholds(mp.false_negatives).flatten()

        # normalize
        support = tp + fn
        with np.errstate(invalid="ignore", divide="ignore"):
            tp_rel = tp / support
            fp_rel = fp / support
            fn_rel = fn / support

            # sort by f1
            sort_scores = 2 * tp / (2 * tp + fp + fn)

        K = len(self.eval_result.mp.cat_names)
        sort_indices = np.argsort(sort_scores)
        cat_names_sorted = [self.eval_result.mp.cat_names[i] for i in sort_indices]
        tp_rel, fn_rel, fp_rel = tp_rel[sort_indices], fn_rel[sort_indices], fp_rel[sort_indices]

        objects_count = np.concatenate([tp[sort_indices], fn[sort_indices], fp[sort_indices]])
        data = {
            "Type": ["TP"] * K + ["FN"] * K + ["FP"] * K,
            "category": cat_names_sorted * 3,
        }
        y_label = ""
        if switch_key == "normalized":
            y_label = "Objects Fraction"
            data["count"] = np.concatenate([tp_rel, fn_rel, fp_rel])
        elif switch_key == "absolute":
            y_label = "Objects Count"
            data["count"] = objects_count

        df = pd.DataFrame(data)

        color_map = {"TP": "#8ACAA1", "FN": "#dd3f3f", "FP": "#F7ADAA"}
        fig = px.bar(
            df,
            x="category",
            y="count",
            color="Type",
            height=500,
            width=1000,
            labels={"count": y_label, "category": "Class"},
            color_discrete_map=color_map,
        )
        xaxis_title = fig.layout.xaxis.title.text
        yaxis_title = fig.layout.yaxis.title.text
        if switch_key == "normalized":

            fig.update_traces(
                hovertemplate="Type=%{fullData.name} <br>"
                + xaxis_title
                + "=%{x}<br>"
                + yaxis_title
                + "=%{y:.2f}<extra></extra>"
            )
        elif switch_key == "absolute":
            fig.update_traces(
                hovertemplate="Type=%{fullData.name} <br>"
                + xaxis_title
                + "=%{x}<br>"
                + yaxis_title
                + "=%{y}<extra></extra>",
            )
        return fig

    def get_click_data(self) -> Dict:
        if not self.clickable:
            return
        res = {}
        res["layoutTemplate"] = [None, None, None]

        res["clickData"] = {}
        for class_name, v1 in self.eval_result.click_data.outcome_counts_by_class.items():
            for outcome, matches_data in v1.items():
                key = class_name + self._keypair_sep + outcome
                res["clickData"][key] = {}
                res["clickData"][key]["imagesIds"] = []
                res["clickData"][key][
                    "title"
                ] = f"Images with objects of class '{class_name}' and outcome '{outcome}'"

                img_ids = set()
                for match_data in matches_data:
                    img_comparison_data = self.eval_result.matched_pair_data[
                        match_data["gt_img_id"]
                    ]
                    if outcome == "FN":
                        img_ids.add(img_comparison_data.diff_image_info.id)
                    else:
                        img_ids.add(img_comparison_data.pred_image_info.id)
                res["clickData"][key]["imagesIds"] = list(img_ids)
                res["clickData"][key]["filters"] = [
                    {
                        "type": "tag",
                        "tagId": "confidence",
                        "value": [self.eval_result.mp.conf_threshold, 1],
                    },
                    {"type": "tag", "tagId": "outcome", "value": outcome},
                ]
        return res



from __future__ import annotations

import os
from pathlib import Path

from supervisely.io.json import dump_json_file
from supervisely.nn.benchmark.object_detection.evaluator import (
    ObjectDetectionEvalResult,
    ObjectDetectionEvaluator,
)
from supervisely.nn.benchmark.object_detection.metric_provider import MetricProvider
from supervisely.nn.benchmark.utils import calculate_metrics, read_coco_datasets


class InstanceSegmentationEvalResult(ObjectDetectionEvalResult):
    mp_cls = MetricProvider
    PRIMARY_METRIC = "mAP"

    @classmethod
    def from_evaluator(
        cls, evaulator: InstanceSegmentationEvaluator
    ) -> InstanceSegmentationEvalResult:
        """Method to customize loading of the evaluation result."""
        eval_result = cls()
        eval_result.eval_data = evaulator.eval_data
        eval_result.coco_gt = evaulator.cocoGt
        eval_result.coco_dt = evaulator.cocoDt
        eval_result._prepare_data()
        return eval_result


class InstanceSegmentationEvaluator(ObjectDetectionEvaluator):
    EVALUATION_PARAMS_YAML_PATH = f"{Path(__file__).parent}/evaluation_params.yaml"
    eval_result_cls = InstanceSegmentationEvalResult
    accepted_shapes = ["polygon", "bitmap"]

    def evaluate(self):
        try:
            self.cocoGt_json, self.cocoDt_json = self._convert_to_coco()
        except AssertionError as e:
            raise ValueError(
                f"{e}. Please make sure that your GT and DT projects are correct. "
                "If GT project has nested datasets and DT project was created with NN app, "
                "try to use newer version of NN app."
            )

        self._dump_datasets()
        self.cocoGt, self.cocoDt = read_coco_datasets(self.cocoGt_json, self.cocoDt_json)
        with self.pbar(message="Evaluation: Calculating metrics", total=5) as p:
            self.eval_data = calculate_metrics(
                self.cocoGt,
                self.cocoDt,
                iouType="segm",
                progress_cb=p.update,
                evaluation_params=self.evaluation_params,
            )
        self._dump_eval_results()

    def _dump_eval_results(self):
        _, _, eval_data_path = self._get_eval_paths()
        self._dump_pickle(self.eval_data, eval_data_path)

    def _get_eval_paths(self):
        base_dir = self.result_dir
        cocoGt_path = os.path.join(base_dir, "cocoGt.json")
        cocoDt_path = os.path.join(base_dir, "cocoDt.json")
        eval_data_path = os.path.join(base_dir, "eval_data.pkl")
        return cocoGt_path, cocoDt_path, eval_data_path

    def _dump_datasets(self):
        cocoGt_path, cocoDt_path, _ = self._get_eval_paths()
        dump_json_file(self.cocoGt_json, cocoGt_path, indent=None)
        dump_json_file(self.cocoDt_json, cocoDt_path, indent=None)



import asyncio
import os
import time

from tqdm import tqdm

from supervisely import Api, logger
from supervisely._utils import run_coroutine
from supervisely.io.fs import clean_dir, ensure_base_path
from supervisely.project.download import download_fast, download_to_cache
from supervisely.project.project import _download_project, _download_project_async

# LOG_LEVEL = "INFO"
LOG_LEVEL = "DEBUG"
PROJECT_ID = 11
DATSET_ID = 12
home_dir = os.path.expanduser("~")
common_path = os.path.join(home_dir, "test_project_download/")
save_path = os.path.join(common_path, "old/")
save_path_async = os.path.join(common_path, "async/")
ensure_base_path(common_path)
api = Api.from_env()

clean_dir(common_path)


def main_df():
    download_fast(api, PROJECT_ID, save_path_async)


def main_dpa():
    coroutine = _download_project_async(api, PROJECT_ID, save_path_async, resume_download=True)
    start = time.monotonic()
    run_coroutine(coroutine)
    finish = time.monotonic() - start
    print(f"Time taken for async method: {finish}")
    print(f"Project downloaded to {save_path_async}")


def main_dptc():
    start = time.monotonic()
    download_to_cache(api, PROJECT_ID)
    finish = time.monotonic() - start
    print(f"Time taken for cache method: {finish}")


def main_dps():
    start = time.monotonic()
    _download_project(api, PROJECT_ID, save_path)
    finish = time.monotonic() - start
    print(f"Time taken for sync method: {finish}")


def compare_downloads():
    main_dps()
    clean_dir(save_path)
    main_dpa()
    clean_dir(save_path_async)


def ann_db():
    imgs = api.image.get_list(DATSET_ID)
    img_ids = [img.id for img in imgs]
    pbar = tqdm(desc="Downloading annotations", unit="B", unit_scale=True)
    coroutne = api.annotation.download_batch_async(
        DATSET_ID,
        img_ids,
        semaphore=asyncio.Semaphore(200),
        progress_cb=pbar,
        progress_cb_type="size",
        # force_metadata_for_links=False,
    )
    start = time.monotonic()
    anns = run_coroutine(coroutne)
    finish = time.monotonic() - start
    print(f"Time taken to download {len(anns)} annotations: {finish}")


if __name__ == "__main__":
    try:
        # api.logger.setLevel(LOG_LEVEL)
        # ann_db()
        # main_dpa()  # to download and save project as files (async)
        # main_dps()  # to download and save project as files (sync)
        # compare_downloads()  # to compare the time taken for downloading and saving project as files (sync vs async)
        main_df()
    except KeyboardInterrupt:
        logger.info("Stopped by user")



from supervisely.cli.teamfiles.teamfiles_remove import remove_file_run, remove_directory_run

from supervisely.cli.teamfiles.teamfiles_upload import upload_directory_run

from supervisely.cli.teamfiles.teamfiles_download import download_directory_run



# coding: utf-8

# fields names
ID = "id"
KEY = "key"
INDEX = "index"
FIGURES = "figures"
OBJECT_ID = "objectId"
OBJECT_KEY = "objectKey"
FRAME_RANGE = "frameRange"

TAGS_MAP = "tags"
OBJECTS_MAP = "objects"
FIGURES_MAP = "figures"
VIDEOS_MAP = "videos"

FRAMES = "frames"

OBJECTS = "objects"
FRAMES_COUNT = "framesCount"
VIDEO_ID = "videoId"
VIDEO_NAME = "videoName"
TAGS = "tags"
DESCRIPTION = "description"

IMG_DESCRIPTION = "description"
IMG_SIZE = "size"
IMG_SIZE_WIDTH = "width"
IMG_SIZE_HEIGHT = "height"

META = "meta"



# coding: utf-8

# docs
from typing import List, Optional, Union

from supervisely.annotation.tag_meta import TagMeta
from supervisely.api.entity_annotation.tag_api import TagApi
from supervisely.api.module_api import ApiField
from supervisely.collection.key_indexed_collection import KeyIndexedCollection
from supervisely.project.project_meta import ProjectMeta
from supervisely.video_annotation.key_id_map import KeyIdMap
from supervisely.video_annotation.video_tag import VideoTag
from supervisely.video_annotation.video_tag_collection import VideoTagCollection


class VideoTagApi(TagApi):
    """
    :class:`VideoTag<supervisely.video_annotation.video_tag.VideoTag>` for a single video. :class:`VideoTagApi<VideoTagApi>` object is immutable.
    """

    _entity_id_field = ApiField.VIDEO_ID
    _method_bulk_add = "videos.tags.bulk.add"

    def remove_from_video(self, tag_id: int) -> None:
        """
        Remove tag from video.

        :param tag_id: VideoTag ID in Supervisely.
        :type tag_id: int
        :return: None
        :rtype: :class:`NoneType`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            api.video.tag.remove_from_video(video_tag_id)

        """

        self._api.post("videos.tags.remove", {ApiField.ID: tag_id})

    def update_frame_range(self, tag_id: int, frame_range: List[int]) -> None:
        """
        Update VideoTag frame range in video.

        :param tag_id: VideoTag ID in Supervisely.
        :type tag_id: int
        :param frame_range: New VideoTag frame range.
        :type frame_range: List[int]
        :return: None
        :rtype: :class:`NoneType`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            new_frame_range = [5, 10]
            api.video.tag.update_frame_range(video_tag_id, new_frame_range)
        """

        self._api.post(
            "videos.tags.update", {ApiField.ID: tag_id, ApiField.FRAME_RANGE: frame_range}
        )

    def update_value(self, tag_id: int, tag_value: Union[str, int]) -> None:
        """
        Update VideoTag value.

        :param tag_id: VideoTag ID in Supervisely.
        :type tag_id: int
        :param tag_value: New VideoTag value.
        :type tag_value: str or int
        :return: None
        :rtype: :class:`NoneType`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            api.video.tag.update_value(video_tag_id, 'new_tag_value')
        """

        self._api.post("videos.tags.update-value", {ApiField.ID: tag_id, ApiField.VALUE: tag_value})

    def add_tag(
        self,
        project_meta_tag_id: int,
        video_id: int,
        value: Optional[Union[str, int]] = None,
        frame_range: Optional[List[int]] = None,
        is_finished: Optional[bool] = True,
        non_final_value: Optional[bool] = False,
    ) -> int:
        """
        Add VideoTag to video.

        :param project_meta_tag_id: TagMeta ID in Supervisely.
        :type project_meta_tag_id: int
        :param video_id: Video ID in Supervidely.
        :type video_id: int
        :param value: New VideoTag value.
        :type value: str or int
        :param frame_range: New VideoTag frame range.
        :type frame_range: List[int]
        :param is_finished: Specify if the tag is finished.
        :type is_finished: bool
        :param non_final_value: Specify if the tag value is non-final.
        :type non_final_value: bool
        :return: None
        :rtype: dict
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            new_frame_range = [5, 10]
            api.video.tag.add_tag(project_meta_tag_id, video_id, 'tag_value', frame_range)
        """

        request_data = {ApiField.TAG_ID: project_meta_tag_id, ApiField.VIDEO_ID: video_id}
        if value:
            request_data[ApiField.VALUE] = value
        if non_final_value is True:
            is_finished = False
        if frame_range:
            if is_finished is False and frame_range[0] != frame_range[1]:
                raise ValueError("Start frame and end frame must be equal for unfinished tags")
            request_data[ApiField.FRAME_RANGE] = frame_range
        request_data[ApiField.IS_FINISHED] = is_finished
        request_data[ApiField.NON_FINAL_VALUE] = non_final_value

        resp = self._api.post("videos.tags.add", request_data)
        # {'imageId': 3267369, 'tagId': 368985, 'id': 2296676, 'createdAt': '2022-09-20T11:52:33.829Z', 'updatedAt': '2022-09-20T11:52:33.829Z', 'labelerLogin': 'max', isFinished: true, nonFinalValue: false}
        return resp.json()

    def add(self, video_id: int, tag: VideoTag, update_id_inplace=True) -> int:
        """
        Add VideoTag to video.

        :param video_id: Video ID in Supervidely.
        :type video_id: int
        :param tag: VideoTag j,ject.
        :type tag: VideoTag
        :param update_id_inplace: Specify if
        :return: VideoTag ID in Supervisely
        :rtype: int
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            video_id = 19402023

            tag_meta = sly.TagMeta('Animal', sly.TagValueType.NONE)
            # Tag has to exists in project.
            tag = VideoTag(tag_meta)

            api.video.tag.add(video_id=video_id, tag=tag)
        """

        from supervisely.project.project_meta import ProjectMeta

        if tag.meta.sly_id is None:
            if update_id_inplace is True:
                video_info = self._api.video.get_info_by_id(video_id)
                meta_json = self._api.project.get_meta(video_info.project_id)
                meta = ProjectMeta.from_json(meta_json)
                server_tag_meta = meta.get_tag_meta(tag.meta.name)
                if server_tag_meta is None:
                    raise KeyError(
                        f"Tag with name {tag.meta.name} not found in project with id {video_info.project_id}"
                    )
                tag.meta._set_id(server_tag_meta.sly_id)
            else:
                raise ValueError("tag_meta.sly_id is None, get updated project meta from server")

        resp_json = self.add_tag(
            tag.meta.sly_id,
            video_id,
            tag.value,
            tag.frame_range,
            tag.is_finished,
            tag.non_final_value,
        )
        tag_id = resp_json.get("id")
        created_at = resp_json.get("createdAt")
        updated_at = resp_json.get("updatedAt")
        user = resp_json.get("labelerLogin")
        if update_id_inplace is True and tag_id is not None:
            tag._set_id(tag_id)
        if update_id_inplace is True and created_at is not None:
            tag._set_created_at(created_at)
        if update_id_inplace is True and updated_at is not None:
            tag._set_updated_at(updated_at)
        if update_id_inplace is True and user is not None:
            tag._set_labeler_login(user)
        return tag_id

    def download_list(self, id: int, project_meta: ProjectMeta) -> VideoTagCollection:
        """
        Download VideoTagCollection with all tags of the video.

        :param id: Video ID in Supervidely.
        :type id: int
        :param project_meta_tag_id: TagMeta ID in Supervisely.
        :type project_meta_tag_id: int
        :return: All tags of the video in VideoTagCollection format
        :rtype: VideoTagCollection
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            video_id = 19402023
            project_id = 17209

            meta_json = api.project.get_meta(project_id)
            project_meta = sly.ProjectMeta.from_json(meta_json)

            tags_collection = api.video.tag.download_list(video_id, project_meta)
        """

        data = self._api.video.get_json_info_by_id(id, True)
        tags_json = data["tags"]
        # for tag_json in tags_json:
        #     tag_meta_id = tag_json["tagId"]
        #     tag_meta = project_meta.tag_metas.get_by_id(tag_meta_id)
        #     if tag_meta is None:
        #         raise KeyError(
        #             f"Tag meta with id={tag_meta_id} not found in project meta. Please, update project meta from server"
        #         )
        #     tag_json["name"] = tag_meta.name
        # tags = VideoTagCollection.from_json(tags_json, project_meta.tag_metas)
        # return tags
        tags = VideoTagCollection.from_api_response(tags_json, project_meta.tag_metas)
        return tags

    def remove(self, tag: VideoTag):
        """
        Remove tag from video.

        :param VideoTag in Supervisely.
        :type tag: VideoTag
        :return: None
        :rtype: :class:`NoneType`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            api.video.tag.remove(video_tag)

        """

        if tag.sly_id is None:
            raise ValueError(
                "Only tags with ID (tag.sly_id field) can be removed. Such tags have to be downloaded from server or have ID"
            )
        self.remove_from_video(tag.sly_id)


class VideoObjectTagApi(TagApi):
    _entity_id_field = ApiField.ENTITY_ID
    _method_bulk_add = "annotation-objects.tags.bulk.add"

    def add(
        self,
        tag_meta_id: int,
        object_id: int,
        value: Optional[Union[str, int]] = None,
        frame_range: Optional[List[int]] = None,
        is_finished: Optional[bool] = True,
        non_final_value: Optional[bool] = False,
    ) -> int:
        """Add a tag to an annotation object.

        :param tag_meta_id: TagMeta ID in project `tag_metas`
        :type tag_meta_id: int
        :param object_id: Object ID in project annotation objects
        :type object_id: int
        :param value: possible_values from TagMeta, defaults to None
        :type value: Optional[Union[str, int]], optional
        :param frame_range: array of strictly 2 frame numbers, defaults to None
        :type frame_range: Optional[List[int]], optional
        :return: ID of the tag assigned to the object
        :rtype: int
        """
        request_body = {
            ApiField.TAG_ID: tag_meta_id,
            ApiField.OBJECT_ID: object_id,
        }
        if value is not None:
            request_body[ApiField.VALUE] = value
        if non_final_value is True:
            is_finished = False
        if frame_range is not None:
            if is_finished is False and frame_range[0] != frame_range[1]:
                raise ValueError("Start frame and end frame must be equal for unfinished tags")
            request_body[ApiField.FRAME_RANGE] = frame_range
        request_body[ApiField.IS_FINISHED] = is_finished
        request_body[ApiField.NON_FINAL_VALUE] = non_final_value

        response = self._api.post("annotation-objects.tags.add", request_body)
        id = response.json()[ApiField.ID]
        return id

    def remove(self, tag_id: int) -> None:
        """Remove tag from video annotation object.

        :param tag_id: tag ID of certain object
        :type tag_id: int
        """
        request_body = {ApiField.ID: tag_id}

        self._api.post("annotation-objects.tags.remove", request_body)

    def update_value(self, tag_id: int, value: Union[str, int]) -> None:
        """Update tag value for video annotation object.
        You could use only those values, which are correspond to TagMeta `value_type` and `possible_values`

        :param tag_id: tag ID of certain object
        :type tag_id: int
        :param value: possible_values from TagMeta
        :type value: Union[str, int]
        """
        request_body = {
            ApiField.ID: tag_id,
            ApiField.VALUE: value,
        }
        self._api.post("annotation-objects.tags.update-value", request_body)

    def update_frame_range(self, tag_id: int, frame_range: List[int]) -> None:
        """Update tag frames for video annotation object.

        :param tag_id: tag ID of certain object
        :type tag_id: int
        :param frame_range: range of possible frames, it must always have strictly 2 values
        :type frame_range: List[int]
        """
        request_body = {
            ApiField.ID: tag_id,
            ApiField.FRAME_RANGE: frame_range,
        }
        self._api.post("annotation-objects.tags.update", request_body)

    def append_to_entity(
        self,
        entity_id: int,
        project_id: int,
        tags: KeyIndexedCollection,
        object_id: int,
        key_id_map: KeyIdMap = None,
    ):
        """
        Add tags to entity in project with given ID.

        :param entity_id: ID of the entity in Supervisely to add a tag to
        :type entity_id: int
        :param project_id: Project ID in Supervisely.
        :type project_id: int
        :param tags: Collection of tags
        :type tags: KeyIndexedCollection
        :param object_id: ID of the object in Supervisely to add a tag to
        :type object_id: int, optional
        :param key_id_map: KeyIdMap object.
        :type key_id_map: KeyIdMap, optional
        :return: List of tags IDs
        :rtype: list
        """

        if len(tags) == 0:
            return []
        tags_json, tags_keys = self._tags_to_json(tags, project_id=project_id)
        tags_to_add = []
        for tag in tags_json:
            if object_id is not None:
                tag[ApiField.OBJECT_ID] = object_id
            tags_to_add.append(tag)
        ids = self._append_json(entity_id, tags_to_add)
        KeyIdMap.add_tags_to(key_id_map, tags_keys, ids)
        return ids



from typing import Dict

from supervisely.nn.benchmark.visualization.widgets.widget import BaseWidget


class NotificationWidget(BaseWidget):

    def __init__(
        self,
        name: str,
        title: str,
        desc: str = None,
    ) -> None:
        super().__init__(name, title=title)
        self.desc = desc

    def save_data(self, basepath: str) -> None:
        return

    def get_state(self) -> Dict:
        return {}

    def to_html(self) -> str:
        res = f"""
                    <div style="margin-top: 20px; margin-bottom: 20px;">
                        <sly-iw-notification              
                        iw-widget-id="{ self.id }"
                        >"""
        if self.title:
            res += f"""
                            <span slot="title">
                                { self.title }
                            </span>"""
        if self.desc:
            res += f"""
                            <span slot="description">
                                { self.desc }
                            </span>"""
        res += """
                        </sly-iw-notification>
                    </div>"""
        return res



"""
GUI module for training application.

This module provides the `TrainGUI` class that handles the graphical user interface (GUI) for managing
training workflows in Supervisely.
"""

from os import environ
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import supervisely.io.env as sly_env
import supervisely.io.json as sly_json
from supervisely import Api, ProjectMeta
from supervisely._utils import is_production
from supervisely.app.widgets import Button, Card, Stepper, Widget
from supervisely.geometry.bitmap import Bitmap
from supervisely.geometry.graph import GraphNodes
from supervisely.geometry.polygon import Polygon
from supervisely.geometry.rectangle import Rectangle
from supervisely.nn.task_type import TaskType
from supervisely.nn.training.gui.classes_selector import ClassesSelector
from supervisely.nn.training.gui.hyperparameters_selector import HyperparametersSelector
from supervisely.nn.training.gui.input_selector import InputSelector
from supervisely.nn.training.gui.model_selector import ModelSelector
from supervisely.nn.training.gui.tags_selector import TagsSelector
from supervisely.nn.training.gui.train_val_splits_selector import TrainValSplitsSelector
from supervisely.nn.training.gui.training_artifacts import TrainingArtifacts
from supervisely.nn.training.gui.training_logs import TrainingLogs
from supervisely.nn.training.gui.training_process import TrainingProcess
from supervisely.nn.training.gui.utils import set_stepper_step, wrap_button_click
from supervisely.nn.utils import ModelSource, RuntimeType


class StepFlow:
    """
    Manages the flow of steps in the GUI, including wrappers and button handlers.

    Allows flexible configuration of dependencies between steps and automatically
    sets up proper handlers based on layout from app_options.
    """

    def __init__(self, stepper: Stepper, app_options: Dict[str, Any]):
        """
        Initializes the step manager.

        :param stepper: Stepper object for step navigation
        :param app_options: Application options
        """
        self.stepper = stepper
        self.app_options = app_options
        self.collapsable = app_options.get("collapsable", False)
        self.steps = {}  # Step configuration
        self.step_sequence = []  # Step sequence

    def register_step(
        self,
        name: str,
        card: Card,
        button: Optional[Button] = None,
        widgets_to_disable: Optional[List[Widget]] = None,
        validation_text: Optional[Widget] = None,
        validation_func: Optional[Callable] = None,
        position: Optional[int] = None,
    ) -> "StepFlow":
        """
        Registers a step in the GUI.

        :param name: Unique step name
        :param card: Step card widget
        :param button: Button for proceeding to the next step (optional)
        :param widgets_to_disable: Widgets to disable during validation (optional)
        :param validation_text: Widget for displaying validation text (optional)
        :param validation_func: Validation function (optional)
        :param position: Step position in the sequence (starting from 0)
        :return: Current StepFlow object for method chaining
        """
        self.steps[name] = {
            "card": card,
            "button": button,
            "widgets_to_disable": widgets_to_disable or [],
            "validation_text": validation_text,
            "validation_func": validation_func,
            "position": position,
            "next_steps": [],
            "on_select_click": [],
            "on_reselect_click": [],
            "wrapper": None,
            "has_button": button is not None,
        }

        if position is not None:
            while len(self.step_sequence) <= position:
                self.step_sequence.append(None)
            self.step_sequence[position] = name

        return self

    def set_next_steps(self, step_name: str, next_steps: List[str]) -> "StepFlow":
        """
        Sets the list of next steps for the given step.

        :param step_name: Current step name
        :param next_steps: List of names of the next steps
        :return: Current StepFlow object for method chaining
        """
        if step_name in self.steps:
            self.steps[step_name]["next_steps"] = next_steps
        return self

    def add_on_select_actions(
        self, step_name: str, actions: List[Callable], is_reselect: bool = False
    ) -> "StepFlow":
        """
        Adds actions to be executed when a step is selected/reselected.

        :param step_name: Step name
        :param actions: List of functions to execute
        :param is_reselect: True if these are actions for reselection, otherwise False
        :return: Current StepFlow object for method chaining
        """
        if step_name in self.steps:
            key = "on_reselect_click" if is_reselect else "on_select_click"
            self.steps[step_name][key].extend(actions)
        return self

    def build_wrappers(self) -> Dict[str, Callable]:
        """
        Creates wrappers for all steps based on established dependencies.

        :return: Dictionary with created wrappers by step name
        """
        valid_sequence = [s for s in self.step_sequence if s is not None and s in self.steps]

        for step_name in reversed(valid_sequence):
            step = self.steps[step_name]

            cards_to_unlock = []
            for next_step_name in step["next_steps"]:
                if next_step_name in self.steps:
                    cards_to_unlock.append(self.steps[next_step_name]["card"])

            callback = None
            if step["next_steps"] and step["has_button"]:
                for next_step_name in step["next_steps"]:
                    if (
                        next_step_name in self.steps
                        and self.steps[next_step_name].get("wrapper")
                        and self.steps[next_step_name]["has_button"]
                    ):
                        callback = self.steps[next_step_name]["wrapper"]
                        break

            if step["has_button"]:
                wrapper = wrap_button_click(
                    button=step["button"],
                    cards_to_unlock=cards_to_unlock,
                    widgets_to_disable=step["widgets_to_disable"],
                    callback=callback,
                    validation_text=step["validation_text"],
                    validation_func=step["validation_func"],
                    on_select_click=step["on_select_click"],
                    on_reselect_click=step["on_reselect_click"],
                    collapse_card=(step["card"], self.collapsable),
                )

                step["wrapper"] = wrapper

        return {
            name: self.steps[name]["wrapper"]
            for name in self.steps
            if self.steps[name].get("wrapper") and self.steps[name]["has_button"]
        }

    def setup_button_handlers(self) -> None:
        """
        Sets up handlers for buttons of all steps.
        """
        positions = {}
        pos = 1

        for i, step_name in enumerate(self.step_sequence):
            if step_name is not None and step_name in self.steps:
                positions[step_name] = pos
                pos += 1

        for step_name, step in self.steps.items():
            if step_name in positions and step.get("wrapper") and step["has_button"]:

                button = step["button"]
                wrapper = step["wrapper"]
                position = positions[step_name]
                next_position = position + 1

                def create_handler(btn, cb, next_pos):
                    def handler():
                        cb()
                        set_stepper_step(self.stepper, btn, next_pos=next_pos)

                    return handler

                button.click(create_handler(button, wrapper, next_position))

    def build(self) -> Dict[str, Callable]:
        """
        Performs the complete setup of the step system.

        :return: Dictionary with created wrappers by step name
        """
        wrappers = self.build_wrappers()
        self.setup_button_handlers()
        return wrappers


class TrainGUI:
    """
    A class representing the GUI for training workflows.

    This class sets up and manages GUI components such as project selection,
    train/validation split selection, model selection, hyperparameters selection,
    and the training process.

    :param framework_name: Name of the ML framework being used.
    :type framework_name: str
    :param models: List of available models.
    :type models: list
    :param hyperparameters: Hyperparameters for training.
    :type hyperparameters: dict
    :param app_options: Application options for customization.
    :type app_options: dict, optional
    """

    def __init__(
        self,
        framework_name: str,
        models: list,
        hyperparameters: dict,
        app_options: dict = None,
    ):
        self._api = Api.from_env()
        if is_production():
            self.task_id = sly_env.task_id()
        else:
            self.task_id = sly_env.task_id(raise_not_found=False)
            if self.task_id is None:
                self.task_id = -1

        self.framework_name = framework_name
        self.models = models
        self.hyperparameters = hyperparameters
        self.app_options = app_options
        self.collapsable = self.app_options.get("collapsable", False)
        self.need_convert_shapes_for_bm = False

        self.team_id = sly_env.team_id(raise_not_found=False)
        self.workspace_id = sly_env.workspace_id(raise_not_found=False)
        self.project_id = sly_env.project_id()
        self.project_info = self._api.project.get_info_by_id(self.project_id)
        if self.project_info.type is None:
            raise ValueError(f"Project with ID: '{self.project_id}' does not exist or was archived")

        self.project_meta = ProjectMeta.from_json(self._api.project.get_meta(self.project_id))

        if self.workspace_id is None:
            self.workspace_id = self.project_info.workspace_id
            environ["WORKSPACE_ID"] = str(self.workspace_id)
        if self.team_id is None:
            self.team_id = self.project_info.team_id
            environ["TEAM_ID"] = str(self.team_id)

        # ---------- Parse selector options ----------
        self._classes_selector_opts = self.app_options.get("classes_selector", {})
        self._tags_selector_opts = self.app_options.get("tags_selector", {})
        self._train_val_splits_opts = self.app_options.get("train_val_splits_selector", {})
        self._model_selector_opts = self.app_options.get("model_selector", {})

        self.show_classes_selector = self._classes_selector_opts.get("enabled", True)
        self.show_tags_selector = self._tags_selector_opts.get("enabled", False)
        self.show_train_val_splits_selector = self._train_val_splits_opts.get("enabled", True)
        self.show_model_selector = self._model_selector_opts.get("enabled", True)

        # Ensure train_val_splits_methods compatibility
        self._train_val_methods = self._train_val_splits_opts.get("methods", [])
        # --------------------------------------------------------- #

        # ------------------------------------------------- #
        self.steps = []

        # 1. Project selection
        self.input_selector = InputSelector(self.project_info, self.app_options)
        self.steps.append(self.input_selector.card)

        # 2. Train/val split
        self.train_val_splits_selector = None
        if self.show_train_val_splits_selector:
            self.train_val_splits_selector = TrainValSplitsSelector(
                self._api, self.project_id, self.app_options
            )
            self.steps.append(self.train_val_splits_selector.card)

        # 3. Select Classes
        self.classes_selector = None
        if self.show_classes_selector:
            self.classes_selector = ClassesSelector(self.project_id, [], self.app_options)
            self.steps.append(self.classes_selector.card)

        # 4. Select Tags
        self.tags_selector = None
        if self.show_tags_selector:
            self.tags_selector = TagsSelector(self.project_id, [], self.app_options)
            self.steps.append(self.tags_selector.card)

        # 5. Model selection
        self.model_selector = ModelSelector(
            self._api, self.framework_name, self.models, self.app_options
        )
        if self.show_model_selector:
            self.steps.append(self.model_selector.card)

        # 6. Training parameters (yaml)
        self.hyperparameters_selector = HyperparametersSelector(
            self.hyperparameters, self.app_options
        )
        self.steps.append(self.hyperparameters_selector.card)

        # 7. Start Training
        self.training_process = TrainingProcess(self.app_options)
        self.steps.append(self.training_process.card)

        # 8. Training logs
        self.training_logs = TrainingLogs(self.app_options)
        self.steps.append(self.training_logs.card)

        # 9. Training Artifacts
        self.training_artifacts = TrainingArtifacts(self._api, self.app_options)
        self.steps.append(self.training_artifacts.card)

        # Stepper layout
        self.stepper = Stepper(widgets=self.steps)
        # ------------------------------------------------- #

        # Button utils
        def disable_hyperparams_editor():
            if self.hyperparameters_selector.editor.readonly:
                self.hyperparameters_selector.editor.readonly = False
            else:
                self.hyperparameters_selector.editor.readonly = True

        def set_experiment_name():
            model_name = self.model_selector.get_model_name()
            if model_name is None:
                experiment_name = "Enter experiment name"
            else:
                if self.task_id == -1:
                    experiment_name = f"debug_{self.project_info.name}_{model_name}"
                else:
                    experiment_name = f"{self.task_id}_{self.project_info.name}_{model_name}"

            if experiment_name == self.training_process.get_experiment_name():
                return
            self.training_process.set_experiment_name(experiment_name)

        def need_convert_class_shapes() -> bool:
            if self.hyperparameters_selector.run_model_benchmark_checkbox is not None:
                if not self.hyperparameters_selector.run_model_benchmark_checkbox.is_checked():
                    self.hyperparameters_selector.model_benchmark_auto_convert_warning.hide()
                    self.need_convert_shapes_for_bm = False
                else:
                    task_type = self.model_selector.get_selected_task_type()

                    def _need_convert(shape):
                        if task_type == TaskType.OBJECT_DETECTION:
                            return shape != Rectangle.geometry_name()
                        elif task_type in [
                            TaskType.INSTANCE_SEGMENTATION,
                            TaskType.SEMANTIC_SEGMENTATION,
                        ]:
                            return shape == Polygon.geometry_name()
                        return False

                    if self.classes_selector is not None:
                        data = self.classes_selector.classes_table._table_data
                        selected_classes = set(
                            self.classes_selector.classes_table.get_selected_classes()
                        )
                        empty = set(
                            r[0]["data"] for r in data if r[2]["data"] == 0 and r[3]["data"] == 0
                        )
                        need_convert = set(
                            r[0]["data"] for r in data if _need_convert(r[1]["data"])
                        )
                    else:
                        # Set project meta classes when classes selector is disabled
                        selected_classes = set(cls.name for cls in self.project_meta.obj_classes)
                        need_convert = set(
                            obj_class.name
                            for obj_class in self.project_meta.obj_classes
                            if _need_convert(obj_class.geometry_type)
                        )
                        empty = set()

                    if need_convert.intersection(selected_classes - empty):
                        self.hyperparameters_selector.model_benchmark_auto_convert_warning.show()
                        self.need_convert_shapes_for_bm = True
                    else:
                        self.hyperparameters_selector.model_benchmark_auto_convert_warning.hide()
                        self.need_convert_shapes_for_bm = False
            else:
                self.need_convert_shapes_for_bm = False

        def validate_class_shape_for_model_task():
            task_type = self.model_selector.get_selected_task_type()
            if self.classes_selector is not None:
                classes = self.classes_selector.get_selected_classes()
            else:
                classes = list(self.project_meta.obj_classes.keys())

            required_geometries = {
                TaskType.INSTANCE_SEGMENTATION: {Polygon, Bitmap},
                TaskType.SEMANTIC_SEGMENTATION: {Polygon, Bitmap},
                TaskType.POSE_ESTIMATION: {GraphNodes},
            }
            task_specific_texts = {
                TaskType.INSTANCE_SEGMENTATION: "Only polygon and bitmap shapes are supported for segmentation task",
                TaskType.SEMANTIC_SEGMENTATION: "Only polygon and bitmap shapes are supported for segmentation task",
                TaskType.POSE_ESTIMATION: "Only keypoint (graph) shape is supported for pose estimation task",
            }

            if task_type not in required_geometries:
                return

            wrong_shape_classes = [
                class_name
                for class_name in classes
                if self.project_meta.get_obj_class(class_name).geometry_type
                not in required_geometries[task_type]
            ]

            if wrong_shape_classes:
                specific_text = task_specific_texts[task_type]
                message_text = f"Model task type is {task_type}. {specific_text}. Selected classes have wrong shapes for the model task: {', '.join(wrong_shape_classes)}"
                self.model_selector.validator_text.set(
                    text=message_text,
                    status="warning",
                )

        # ------------------------------------------------- #

        self.step_flow = StepFlow(self.stepper, self.app_options)
        position = 0

        # 1. Input selector
        self.step_flow.register_step(
            "input_selector",
            self.input_selector.card,
            self.input_selector.button,
            self.input_selector.widgets_to_disable,
            self.input_selector.validator_text,
            self.input_selector.validate_step,
            position=position,
        )
        position += 1

        # 2. Train/Val splits selector
        if self.show_train_val_splits_selector and self.train_val_splits_selector is not None:
            self.step_flow.register_step(
                "train_val_splits",
                self.train_val_splits_selector.card,
                self.train_val_splits_selector.button,
                self.train_val_splits_selector.widgets_to_disable,
                self.train_val_splits_selector.validator_text,
                self.train_val_splits_selector.validate_step,
                position=position,
            )
            position += 1

        # 3. Classes selector
        if self.show_classes_selector and self.classes_selector is not None:
            self.step_flow.register_step(
                "classes_selector",
                self.classes_selector.card,
                self.classes_selector.button,
                self.classes_selector.widgets_to_disable,
                self.classes_selector.validator_text,
                self.classes_selector.validate_step,
                position=position,
            )
            position += 1

        # 4. Tags selector
        if self.show_tags_selector and self.tags_selector is not None:
            self.step_flow.register_step(
                "tags_selector",
                self.tags_selector.card,
                self.tags_selector.button,
                self.tags_selector.widgets_to_disable,
                self.tags_selector.validator_text,
                self.tags_selector.validate_step,
                position=position,
            )
            position += 1

        # 5. Model selector
        if self.show_model_selector:
            self.step_flow.register_step(
                "model_selector",
                self.model_selector.card,
                self.model_selector.button,
                self.model_selector.widgets_to_disable,
                self.model_selector.validator_text,
                self.model_selector.validate_step,
                position=position,
            ).add_on_select_actions(
                "model_selector",
                [
                    set_experiment_name,
                    need_convert_class_shapes,
                    validate_class_shape_for_model_task,
                ],
            )
            position += 1

        # 6. Hyperparameters selector
        self.step_flow.register_step(
            "hyperparameters_selector",
            self.hyperparameters_selector.card,
            self.hyperparameters_selector.button,
            self.hyperparameters_selector.widgets_to_disable,
            self.hyperparameters_selector.validator_text,
            self.hyperparameters_selector.validate_step,
            position=position,
        ).add_on_select_actions("hyperparameters_selector", [disable_hyperparams_editor])
        self.step_flow.add_on_select_actions(
            "hyperparameters_selector", [disable_hyperparams_editor], is_reselect=True
        )
        position += 1

        # 7. Training process
        self.step_flow.register_step(
            "training_process",
            self.training_process.card,
            None,
            self.training_process.widgets_to_disable,
            self.training_process.validator_text,
            self.training_process.validate_step,
            position=position,
        )
        position += 1

        # 8. Training logs
        self.step_flow.register_step(
            "training_logs",
            self.training_logs.card,
            None,
            self.training_logs.widgets_to_disable,
            self.training_logs.validator_text,
            self.training_logs.validate_step,
            position=position,
        )
        position += 1

        # 9. Training artifacts
        self.step_flow.register_step(
            "training_artifacts",
            self.training_artifacts.card,
            None,
            self.training_artifacts.widgets_to_disable,
            self.training_artifacts.validator_text,
            self.training_artifacts.validate_step,
            position=position,
        )

        # Set dependencies between steps
        has_train_val_splits = (
            self.show_train_val_splits_selector and self.train_val_splits_selector is not None
        )
        has_classes_selector = self.show_classes_selector and self.classes_selector is not None
        has_tags_selector = self.show_tags_selector and self.tags_selector is not None

        # Set step dependency chain
        # 1. Input selector
        prev_step = "input_selector"
        if has_train_val_splits:
            self.step_flow.set_next_steps(prev_step, ["train_val_splits"])
            prev_step = "train_val_splits"
        if has_classes_selector:
            self.step_flow.set_next_steps(prev_step, ["classes_selector"])
            prev_step = "classes_selector"
        if has_tags_selector:
            self.step_flow.set_next_steps(prev_step, ["tags_selector"])
            prev_step = "tags_selector"

        if self.show_model_selector and self.model_selector is not None:
            self.step_flow.set_next_steps(prev_step, ["model_selector"])
            # Model selector -> hyperparameters
            self.step_flow.set_next_steps("model_selector", ["hyperparameters_selector"])
            prev_step = "model_selector"
        else:
            self.step_flow.set_next_steps(prev_step, ["hyperparameters_selector"])

        # 6. Hyperparameters selector -> 7. Training process
        self.step_flow.set_next_steps("hyperparameters_selector", ["training_process"])

        # 7. Training process -> 8. Training logs
        self.step_flow.set_next_steps("training_process", ["training_logs"])

        # 8. Training logs -> 9. Training artifacts
        self.step_flow.set_next_steps("training_logs", ["training_artifacts"])
        # ------------------------------------------------- #

        # Create all wrappers and set button handlers
        wrappers = self.step_flow.build()

        self.input_selector_cb = wrappers.get("input_selector")
        self.train_val_splits_selector_cb = wrappers.get("train_val_splits")
        self.classes_selector_cb = wrappers.get("classes_selector")
        self.tags_selector_cb = wrappers.get("tags_selector")
        self.model_selector_cb = wrappers.get("model_selector")
        self.hyperparameters_selector_cb = wrappers.get("hyperparameters_selector")
        self.training_process_cb = wrappers.get("training_process")
        self.training_logs_cb = wrappers.get("training_logs")
        self.training_artifacts_cb = wrappers.get("training_artifacts")
        # ------------------------------------------------- #

        # Other handlers
        if self.app_options.get("show_logs_in_gui", False):

            @self.training_logs.logs_button.click
            def show_logs():
                self.training_logs.toggle_logs()

        if self.hyperparameters_selector.run_model_benchmark_checkbox is not None:

            @self.hyperparameters_selector.run_model_benchmark_checkbox.value_changed
            def show_mb_speedtest(is_checked: bool):
                self.hyperparameters_selector.toggle_mb_speedtest(is_checked)
                need_convert_class_shapes()

        # ------------------------------------------------- #

        self.layout: Widget = self.stepper

        # (дублирующийся блок был перемещён выше и здесь удалён)

    def set_next_step(self):
        current_step = self.stepper.get_active_step()
        self.stepper.set_active_step(current_step + 1)

    def set_previous_step(self):
        current_step = self.stepper.get_active_step()
        self.stepper.set_active_step(current_step - 1)

    def set_first_step(self):
        self.stepper.set_active_step(1)

    def set_last_step(self):
        self.stepper.set_active_step(len(self.steps))

    def enable_select_buttons(self):
        """
        Makes all select buttons in the GUI available for interaction.
        """
        if self.input_selector is not None:
            self.input_selector.button.enable()
        if self.train_val_splits_selector is not None:
            self.train_val_splits_selector.button.enable()
        if self.classes_selector is not None:
            self.classes_selector.button.enable()
        if self.tags_selector is not None:
            self.tags_selector.button.enable()
        if self.model_selector is not None:
            self.model_selector.button.enable()
        if self.hyperparameters_selector is not None:
            self.hyperparameters_selector.button.enable()

    def disable_select_buttons(self):
        """
        Makes all select buttons in the GUI unavailable for interaction.
        """
        if self.input_selector is not None:
            self.input_selector.button.disable()
        if self.train_val_splits_selector is not None:
            self.train_val_splits_selector.button.disable()
        if self.classes_selector is not None:
            self.classes_selector.button.disable()
        if self.tags_selector is not None:
            self.tags_selector.button.disable()
        if self.model_selector is not None:
            self.model_selector.button.disable()
        if self.hyperparameters_selector is not None:
            self.hyperparameters_selector.button.disable()

    # Set GUI from config
    def validate_app_state(self, app_state: dict) -> dict:
        """
        Validate the app state dictionary.

        :param app_state: The app state dictionary.
        :type app_state: dict
        """
        if not isinstance(app_state, dict):
            raise ValueError("app_state must be a dictionary")

        show_train_val = self.show_train_val_splits_selector
        show_classes = self.show_classes_selector
        show_tags = self.show_tags_selector

        # Basic required keys always needed
        base_required = {
            "model": ["source"],
            "hyperparameters": (dict, str),
        }
        if show_train_val:
            base_required["train_val_split"] = ["method"]
        if show_classes:
            base_required["classes"] = list
        if show_tags:
            base_required["tags"] = list

        for key, subkeys_or_type in base_required.items():
            if key not in app_state:
                raise KeyError(f"Missing required key in app_state: {key}")
            if isinstance(subkeys_or_type, list):
                for subkey in subkeys_or_type:
                    if subkey not in app_state[key]:
                        raise KeyError(f"Missing required key in app_state['{key}']: {subkey}")
            elif not isinstance(app_state[key], subkeys_or_type):
                valid_types = ""
                if isinstance(subkeys_or_type, tuple):
                    type_names = []
                    for t in subkeys_or_type:
                        if hasattr(t, "__name__"):
                            type_names.append(t.__name__)
                        else:
                            type_names.append(type(t).__name__)
                    valid_types = " or ".join(type_names)
                else:
                    if hasattr(subkeys_or_type, "__name__"):
                        valid_types = subkeys_or_type.__name__
                    else:
                        valid_types = type(subkeys_or_type).__name__
                raise ValueError(f"app_state['{key}'] must be of type {valid_types}")

        # Provide defaults for optional sections when selectors are disabled
        if not show_train_val:
            app_state.setdefault("train_val_split", {"method": "random"})
        if not show_classes:
            app_state.setdefault("classes", [])
        if not show_tags:
            app_state.setdefault("tags", [])

        model = app_state["model"]
        if model["source"] == "Pretrained models":
            if "model_name" not in model:
                raise KeyError("Missing required key in app_state['model']: model_name")
        elif model["source"] == "Custom models":
            custom_keys = ["task_id", "checkpoint"]
            for key in custom_keys:
                if key not in model:
                    raise KeyError(f"Missing required key in app_state['model']: {key}")

        options = app_state.setdefault(
            "options",
            {
                "model_benchmark": {
                    "enable": True,
                    "speed_test": True,
                },
                "cache_project": True,
            },
        )

        if not isinstance(options, dict):
            raise ValueError("app_state['options'] must be a dictionary")

        model_benchmark = options.setdefault(
            "model_benchmark", {"enable": True, "speed_test": True}
        )
        if not isinstance(model_benchmark, dict):
            raise ValueError("app_state['options']['model_benchmark'] must be a dictionary")
        model_benchmark.setdefault("enable", True)
        model_benchmark.setdefault("speed_test", True)

        if not isinstance(options.get("cache_project"), bool):
            raise ValueError("app_state['options']['cache_project'] must be a boolean")

        # Check train val splits
        train_val_splits_settings = app_state.get("train_val_split")
        if train_val_splits_settings.get("method") == "datasets":
            dataset_ids = []
            for parents, dataset in self._api.dataset.tree(self.project_id):
                dataset_ids.append(dataset.id)

            train_datasets = train_val_splits_settings.get("train_datasets", [])
            val_datasets = train_val_splits_settings.get("val_datasets", [])

            missing_datasets_ids = []
            for ds_id in train_datasets + val_datasets:
                if ds_id not in dataset_ids:
                    missing_datasets_ids.append(ds_id)

            if len(missing_datasets_ids) > 0:
                missing_datasets_text = ", ".join([str(ds_id) for ds_id in missing_datasets_ids])
                raise ValueError(
                    f"Datasets with ids: {missing_datasets_text} not found in the project"
                )
        elif train_val_splits_settings.get("method") == "tags":
            train_tag = train_val_splits_settings.get("train_tag")
            val_tag = train_val_splits_settings.get("val_tag")
            if not train_tag or not val_tag:
                raise ValueError("train_tag and val_tag must be specified in tags split method")
        elif train_val_splits_settings.get("method") == "random":
            split = train_val_splits_settings.get("split")
            percent = train_val_splits_settings.get("percent")
            if split not in ["train", "val"]:
                raise ValueError("split must be 'train' or 'val'")
            if not isinstance(percent, int) or not 0 < percent < 100:
                raise ValueError("percent must be an integer in range 1 to 99")
        return app_state

    def load_from_app_state(self, app_state: Union[str, dict]) -> None:
        """
        Load the GUI state from app state dictionary.

        :param app_state: The state dictionary.
        :type app_state: dict

        app_state example:

            app_state = {
                "train_val_split": {
                    "method": "random",
                    "split": "train",
                    "percent": 90
                },
                "classes": ["apple"],
                "model": {
                    "source": "Pretrained models",
                    "model_name": "rtdetr_r50vd_coco_objects365"
                },
                "hyperparameters": hyperparameters, # yaml string
                "options": {
                    "model_benchmark": {
                        "enable": True,
                        "speed_test": True
                    },
                    "cache_project": True,
                "export": {
                    "enable": True,
                    "ONNXRuntime": True,
                    "TensorRT": True
                    },
                }
            }
        """
        if isinstance(app_state, str):
            app_state = sly_json.load_json_file(app_state)
        app_state = self.validate_app_state(app_state)

        options = app_state.get("options", {})
        input_settings = app_state.get("input")
        train_val_splits_settings = app_state.get("train_val_split", {})
        classes_settings = app_state.get("classes", [])
        tags_settings = app_state.get("tags", [])
        model_settings = app_state["model"]
        hyperparameters_settings = app_state["hyperparameters"]

        self._init_input(input_settings, options)
        self._init_train_val_splits(train_val_splits_settings, options)
        self._init_classes(classes_settings, options)
        self._init_tags(tags_settings, options)
        self._init_model(model_settings, options)
        self._init_hyperparameters(hyperparameters_settings, options)

    def _init_input(self, input_settings: Union[dict, None], options: dict) -> None:
        """
        Initialize the input selector with the given settings.

        :param input_settings: The input settings.
        :type input_settings: dict
        :param options: The application options.
        :type options: dict
        """
        # Set Input
        self.input_selector.set_cache(options.get("cache_project", True))
        self.input_selector_cb()
        # ----------------------------------------- #

    def _init_train_val_splits(self, train_val_splits_settings: dict, options: dict) -> None:
        """
        Initialize the train/val splits selector with the given settings.

        :param train_val_splits_settings: The train/val splits settings.
        :type train_val_splits_settings: dict
        :param options: The application options.
        :type options: dict
        """
        if self.train_val_splits_selector is None:
            return  # Selector disabled by app options

        if train_val_splits_settings == {}:
            available_methods = self.app_options.get("train_val_splits_methods", [])
            if available_methods == []:
                method = "random"
                train_val_splits_settings = {"method": method, "split": "train", "percent": 80}
            else:
                method = available_methods[0]
                if method == "random":
                    train_val_splits_settings = {"method": method, "split": "train", "percent": 80}
                elif method == "tags":
                    train_val_splits_settings = {
                        "method": method,
                        "train_tag": "train",
                        "val_tag": "val",
                        "untagged_action": "ignore",
                    }
                elif method == "datasets":
                    train_val_splits_settings = {
                        "method": method,
                        "train_datasets": [],
                        "val_datasets": [],
                    }

        split_method = train_val_splits_settings["method"]
        if split_method == "random":
            split = train_val_splits_settings["split"]
            percent = train_val_splits_settings["percent"]
            self.train_val_splits_selector.train_val_splits.set_random_splits(split, percent)
        elif split_method == "tags":
            train_tag = train_val_splits_settings["train_tag"]
            val_tag = train_val_splits_settings["val_tag"]
            untagged_action = train_val_splits_settings["untagged_action"]
            self.train_val_splits_selector.train_val_splits.set_tags_splits(
                train_tag, val_tag, untagged_action
            )
        elif split_method == "datasets":
            train_datasets = train_val_splits_settings["train_datasets"]
            val_datasets = train_val_splits_settings["val_datasets"]
            self.train_val_splits_selector.train_val_splits.set_datasets_splits(
                train_datasets, val_datasets
            )
        self.train_val_splits_selector_cb()

    def _init_classes(self, classes_settings: list, options: dict) -> None:
        """
        Initialize the classes selector with the given settings.

        :param classes_settings: The classes settings.
        :type classes_settings: list
        :param options: The application options.
        :type options: dict
        """
        if self.classes_selector is None:
            return  # Selector disabled by app options

        # Set Classes
        self.classes_selector.set_classes(classes_settings)
        self.classes_selector_cb()
        # ----------------------------------------- #

    def _init_tags(self, tags_settings: list, options: dict) -> None:
        """
        Initialize the tags selector with the given settings.

        :param tags_settings: The tags settings.
        :type tags_settings: list
        :param options: The application options.
        :type options: dict
        """
        if self.tags_selector is None:
            return  # Selector disabled by app options

        # Set Tags
        self.tags_selector.set_tags(tags_settings)
        self.tags_selector_cb()
        # ----------------------------------------- #

    def _init_model(self, model_settings: dict, options: dict) -> None:
        """
        Initialize the model selector with the given settings.

        :param model_settings: The model settings.
        :type model_settings: dict
        :param options: The application options.
        :type options: dict
        """

        # Pretrained
        if model_settings["source"] == ModelSource.PRETRAINED:
            self.model_selector.model_source_tabs.set_active_tab(ModelSource.PRETRAINED)
            self.model_selector.pretrained_models_table.set_by_model_name(
                model_settings["model_name"]
            )

        # Custom
        elif model_settings["source"] == ModelSource.CUSTOM:
            self.model_selector.model_source_tabs.set_active_tab(ModelSource.CUSTOM)
            self.model_selector.experiment_selector.set_by_task_id(model_settings["task_id"])
            active_row = self.model_selector.experiment_selector.get_selected_row()
            if model_settings["checkpoint"] not in active_row.checkpoints_names:
                raise ValueError(
                    f"Checkpoint '{model_settings['checkpoint']}' not found in selected task"
                )

            active_row.set_selected_checkpoint_by_name(model_settings["checkpoint"])
        self.model_selector_cb()
        # ----------------------------------------- #

    def _init_hyperparameters(self, hyperparameters_settings: dict, options: dict) -> None:
        """
        Initialize the hyperparameters selector with the given settings.

        :param hyperparameters_settings: The hyperparameters settings.
        :type hyperparameters_settings: dict
        :param options: The application options.
        :type options: dict
        """
        self.hyperparameters_selector.set_hyperparameters(hyperparameters_settings)

        model_benchmark_settings = options.get("model_benchmark", None)
        if model_benchmark_settings is not None:
            self.hyperparameters_selector.set_model_benchmark_checkbox_value(
                model_benchmark_settings["enable"]
            )
            self.hyperparameters_selector.set_speedtest_checkbox_value(
                model_benchmark_settings["speed_test"]
                if model_benchmark_settings["enable"]
                else False
            )
        export_weights_settings = options.get("export", None)
        if export_weights_settings is not None:
            self.hyperparameters_selector.set_export_onnx_checkbox_value(
                export_weights_settings.get(RuntimeType.ONNXRUNTIME, False)
            )
            self.hyperparameters_selector.set_export_tensorrt_checkbox_value(
                export_weights_settings.get(RuntimeType.TENSORRT, False)
            )
        self.hyperparameters_selector_cb()

    # ----------------------------------------- #



# coding: utf-8
"""deploy pretrained and custom models"""

from __future__ import annotations

import time
from dataclasses import asdict
from pathlib import Path
from typing import Any, Dict, Literal, Optional, Tuple, Union

import supervisely.io.env as env
from supervisely._utils import get_valid_kwargs
from supervisely.api.api import Api
from supervisely.io.fs import get_file_name_with_ext
from supervisely.nn.experiments import ExperimentInfo
from supervisely.nn.utils import RuntimeType
from supervisely.sly_logger import logger


def get_runtime(runtime: Optional[str] = None):
    from supervisely.nn.utils import RuntimeType

    if runtime is None:
        return None
    aliases = {
        str(RuntimeType.PYTORCH): RuntimeType.PYTORCH,
        str(RuntimeType.ONNXRUNTIME): RuntimeType.ONNXRUNTIME,
        str(RuntimeType.TENSORRT): RuntimeType.TENSORRT,
        "pytorch": RuntimeType.PYTORCH,
        "torch": RuntimeType.PYTORCH,
        "pt": RuntimeType.PYTORCH,
        "onnxruntime": RuntimeType.ONNXRUNTIME,
        "onnx": RuntimeType.ONNXRUNTIME,
        "tensorrt": RuntimeType.TENSORRT,
        "trt": RuntimeType.TENSORRT,
        "engine": RuntimeType.TENSORRT,
    }
    if runtime in aliases:
        return aliases[runtime]
    runtime = aliases.get(runtime.lower(), None)
    if runtime is None:
        raise ValueError(
            f"Runtime '{runtime}' is not supported. Supported runtimes are: {', '.join(aliases.keys())}"
        )
    return runtime


class DeployApi:
    """ """

    def __init__(self, api: "Api"):
        self._api = api

    def load_pretrained_model(
        self,
        session_id: int,
        model_name: str,
        device: Optional[str] = None,
        runtime: str = None,
    ):
        """
        Load a pretrained model in running serving App.

        :param session_id: Task ID of the serving App.
        :type session_id: int
        :param model_name: Model name to deploy.
        :type model_name: str
        :param device: Device string. If not provided, will be chosen automatically.
        :type device: Optional[str]
        :param runtime: Runtime string, if not present will be defined automatically.
        :type runtime: Optional[str]
        """
        from supervisely.nn.utils import ModelSource

        runtime = get_runtime(runtime)
        deploy_params = {}
        deploy_params["model_source"] = ModelSource.PRETRAINED
        deploy_params["device"] = device
        deploy_params["runtime"] = runtime
        self._load_model_from_api(session_id, deploy_params, model_name=model_name)

    def load_custom_model(
        self,
        session_id: int,
        team_id: int,
        artifacts_dir: str,
        checkpoint_name: Optional[str] = None,
        device: Optional[str] = None,
        runtime: str = None,
    ):
        """
        Load a custom model in running serving App.

        :param session_id: Task ID of the serving App.
        :type session_id: int
        :param team_id: Team ID in Supervisely.
        :type team_id: int
        :param artifacts_dir: Path to the artifacts directory in the team fies.
        :type artifacts_dir: str
        :param checkpoint_name: Checkpoint name (with file extension) to deploy, e.g. "best.pt".
            If not provided, checkpoint will be chosen automatically, depending on the app version.
        :type checkpoint_name: Optional[str]
        :param device: Device string. If not provided, will be chosen automatically.
        :type device: Optional[str]
        :param runtime: Runtime string, if not present will be defined automatically.
        :type runtime: Optional[str]
        """
        from supervisely.nn.utils import ModelSource

        runtime = get_runtime(runtime)

        # Train V1 logic (if artifacts_dir does not start with '/experiments')
        if not artifacts_dir.startswith("/experiments"):
            logger.debug("Deploying model from Train V1 artifacts")
            _, _, deploy_params = self._deploy_params_v1(
                team_id, artifacts_dir, checkpoint_name, device, runtime, with_module=False
            )
        else:  # Train V2 logic (when artifacts_dir starts with '/experiments')
            logger.debug("Deploying model from Train V2 artifacts")

            _, _, deploy_params = self._deploy_params_v2(
                team_id, artifacts_dir, checkpoint_name, device, runtime, with_module=False
            )
        deploy_params["model_source"] = ModelSource.CUSTOM
        self._load_model_from_api(session_id, deploy_params)

    def load_custom_model_from_experiment_info(
        self,
        session_id: int,
        experiment_info: "ExperimentInfo",
        checkpoint_name: Optional[str] = None,
        device: Optional[str] = None,
        runtime: str = None,
    ):
        """
        Load a custom model in running serving App based on the training session.

        :param session_id: Task ID of the serving App.
        :type session_id: int
        :param experiment_info: an :class:`ExperimentInfo` object.
        :type experiment_info: ExperimentInfo
        :param checkpoint_name: Checkpoint name (with file extension) to deploy, e.g. "best.pt".
            If not provided, checkpoint will be chosen automatically, depending on the app version.
        :type checkpoint_name: Optional[str]
        :param device: Device string. If not provided, will be chosen automatically.
        :type device: Optional[str]
        :param runtime: Runtime string, if not present will be defined automatically.
        :type runtime: Optional[str]
        """
        from supervisely.nn.utils import ModelSource

        runtime = get_runtime(runtime)
        if checkpoint_name is None:
            checkpoint_name = experiment_info.best_checkpoint
        deploy_params = {
            "device": device,
            "model_source": ModelSource.CUSTOM,
            "model_files": {
                "checkpoint": Path(
                    experiment_info.artifacts_dir, "checkpoints", checkpoint_name
                ).as_posix(),
                "config": Path(
                    experiment_info.artifacts_dir, experiment_info.model_files["config"]
                ).as_posix(),
            },
            "model_info": experiment_info.to_json(),
            "runtime": runtime,
        }
        self._load_model_from_api(session_id, deploy_params)

    def _find_agent(self, team_id: int = None, public=True, gpu=True):
        """
        Find an agent in Supervisely with most available memory.

        :param team_id: Team ID. If not provided, will be taken from the current context.
        :type team_id: Optional[int]
        :param public: If True, can find a public agent.
        :type public: bool
        :param gpu: If True, find an agent with GPU.
        :type gpu: bool
        :return: Agent ID
        :rtype: int
        """
        if team_id is None:
            team_id = env.team_id()
        agents = self._api.agent.get_list_available(team_id, show_public=public, has_gpu=gpu)
        if len(agents) == 0:
            raise ValueError("No available agents found.")
        agent_id_memory_map = {}
        kubernetes_agents = []
        for agent in agents:
            if agent.type == "sly_agent":
                # No multi-gpu support, always take the first one
                agent_id_memory_map[agent.id] = agent.gpu_info["device_memory"][0]["available"]
            elif agent.type == "kubernetes":
                kubernetes_agents.append(agent.id)
        if len(agent_id_memory_map) > 0:
            return max(agent_id_memory_map, key=agent_id_memory_map.get)
        if len(kubernetes_agents) > 0:
            return kubernetes_agents[0]

    def deploy_pretrained_model(
        self,
        framework: Union[str, int],
        model_name: str,
        device: Optional[str] = None,
        runtime: str = None,
        workspace_id: int = None,
        agent_id: Optional[int] = None,
        app: Union[str, int] = None,
        **kwargs,
    ) -> Dict[str, Any]:
        """
        Deploy a pretrained model.

        :param framework: Framework name or Framework ID in Supervisely.
        :type framework: Union[str, int]
        :param model_name: Model name to deploy.
        :type model_name: str
        :param device: Device string. If not provided, will be chosen automatically.
        :type device: Optional[str]
        :param runtime: Runtime string, if not present will be defined automatically.
        :type runtime: Optional[str]
        :param workspace_id: Workspace ID where the app will be deployed. If not provided, will be taken from the current context.
        :type workspace_id: Optional[int]
        :param agent_id: Agent ID. If not provided, will be found automatically.
        :type agent_id: Optional[int]
        :param app: App name or App module ID in Supervisely.
        :type app: Union[str, int]
        :param kwargs: Additional parameters to start the task. See Api.task.start() for more details.
        :type kwargs: Dict[str, Any]
        :return: Task Info
        :rtype: Dict[str, Any]
        :raises ValueError: if no serving apps found for the app name or multiple serving apps found for the app name.
        """
        from supervisely.nn.artifacts import (
            RITM,
            RTDETR,
            Detectron2,
            MMClassification,
            MMDetection,
            MMDetection3,
            MMSegmentation,
            UNet,
            YOLOv5,
            YOLOv5v2,
            YOLOv8,
        )
        from supervisely.nn.artifacts.artifacts import BaseTrainArtifacts
        from supervisely.nn.utils import ModelSource

        workspace_info = self._api.workspace.get_info_by_id(workspace_id)
        if workspace_info is None:
            raise ValueError(f"Workspace with ID {workspace_id} not found")
        team_id = workspace_info.team_id

        # @TODO: Fix debug logs/ Fix code
        # Skip HTTPS redirect check on API init: False. ENV: False. Checked servers: set()
        frameworks_v1 = {
            RITM(team_id).framework_name: RITM(team_id).serve_slug,
            RTDETR(team_id).framework_name: RTDETR(team_id).serve_slug,
            Detectron2(team_id).framework_name: Detectron2(team_id).serve_slug,
            MMClassification(team_id).framework_name: MMClassification(team_id).serve_slug,
            MMDetection(team_id).framework_name: MMDetection(team_id).serve_slug,
            MMDetection3(team_id).framework_name: MMDetection3(team_id).serve_slug,
            MMSegmentation(team_id).framework_name: MMSegmentation(team_id).serve_slug,
            UNet(team_id).framework_name: UNet(team_id).serve_slug,
            YOLOv5(team_id).framework_name: YOLOv5(team_id).serve_slug,
            YOLOv5v2(team_id).framework_name: YOLOv5v2(team_id).serve_slug,
            YOLOv8(team_id).framework_name: YOLOv8(team_id).serve_slug,
        }
        if framework in frameworks_v1:
            slug = frameworks_v1[framework]
            module_id = self.find_serving_app_by_slug(slug)
        else:
            module_id = None
            if isinstance(app, int):
                module_id = app
            elif isinstance(app, str):
                module_id = self._api.app.find_module_id_by_app_name(app)
            else:
                module_id = self.find_serving_app_by_framework(framework)["id"]
        if module_id is None:
            raise ValueError(
                f"Serving app for framework '{framework}' not found. Make sure that you used correct framework name."
            )

        runtime = get_runtime(runtime)
        if agent_id is None:
            agent_id = self._find_agent()

        task_info = self._run_serve_app(agent_id, module_id, workspace_id=workspace_id, **kwargs)
        self.load_pretrained_model(
            task_info["id"], model_name=model_name, device=device, runtime=runtime
        )
        return task_info

    def _find_team_by_path(self, path: str, team_id: int = None, raise_not_found=True):
        if team_id is not None:
            if self._api.file.exists(team_id, path) or self._api.file.dir_exists(
                team_id, path, recursive=False
            ):
                return team_id
            elif raise_not_found:
                raise ValueError(f"Checkpoint '{path}' not found in team provided team")
            else:
                return None
        team_id = env.team_id(raise_not_found=False)
        if team_id is not None:
            if self._api.file.exists(team_id, path) or self._api.file.dir_exists(
                team_id, path, recursive=False
            ):
                return team_id
        teams = self._api.team.get_list()
        team_id = None
        for team in teams:
            if self._api.file.exists(team.id, path):
                if team_id is not None:
                    raise ValueError("Multiple teams have the same checkpoint")
                team_id = team.id
        if team_id is None:
            if raise_not_found:
                raise ValueError("Checkpoint not found")
            else:
                return None
        return team_id

    def deploy_custom_model_by_checkpoint(
        self,
        checkpoint: str,
        device: Optional[str] = None,
        runtime: str = None,
        timeout: int = 100,
        team_id: int = None,
        workspace_id: int = None,
        agent_id: int = None,
        **kwargs,
    ) -> Dict[str, Any]:
        """
        Deploy a custom model based on the checkpoint path.

        :param checkpoint: Path to the checkpoint in Team Files.
        :type checkpoint: str
        :param device: Device string. If not provided, will be chosen automatically.
        :type device: Optional[str]
        :param runtime: Runtime string, if not present will be defined automatically.
        :type runtime: Optional[str]
        :param timeout: Timeout in seconds (default is 100). The maximum time to wait for the serving app to be ready.
        :type timeout: Optional[int]
        :param team_id: Team ID where the artifacts are stored. If not provided, will be taken from the current context.
        :type team_id: Optional[int]
        :param workspace_id: Workspace ID where the app will be deployed. If not provided, will be taken from the current context.
        :type workspace_id: Optional[int]
        :param agent_id: Agent ID. If not provided, will be found automatically.
        :type agent_id: Optional[int]
        :param kwargs: Additional parameters to start the task. See Api.task.start() for more details.
        :type kwargs: Dict[str, Any]
        :return: Task Info
        :rtype: Dict[str, Any]
        :raises ValueError: if validations fail.
        """
        artifacts_dir, checkpoint_name = self._get_artifacts_dir_and_checkpoint_name(checkpoint)
        return self.deploy_custom_model_by_artifacts_dir(
            artifacts_dir=artifacts_dir,
            checkpoint_name=checkpoint_name,
            device=device,
            runtime=runtime,
            timeout=timeout,
            team_id=team_id,
            workspace_id=workspace_id,
            agent_id=agent_id,
            **kwargs,
        )

    def deploy_custom_model_by_artifacts_dir(
        self,
        artifacts_dir: str,
        checkpoint_name: Optional[str] = None,
        device: Optional[str] = None,
        runtime: str = None,
        timeout: int = 100,
        team_id: int = None,
        workspace_id: int = None,
        agent_id: int = None,
        **kwargs,
    ) -> Dict[str, Any]:
        """
        Deploy a custom model based on the artifacts directory.

        :param artifacts_dir: Path to the artifacts directory in the team fies.
        :type artifacts_dir: str
        :param checkpoint_name: Checkpoint name (with file extension) to deploy, e.g. "best.pt".
            If not provided, checkpoint will be chosen automatically, depending on the app version.
        :type checkpoint_name: Optional[str]
        :param device: Device string. If not provided, will be chosen automatically.
        :type device: Optional[str]
        :param runtime: Runtime string, if not present will be defined automatically.
        :type runtime: Optional[str]
        :param timeout: Timeout in seconds (default is 100). The maximum time to wait for the serving app to be ready.
        :type timeout: Optional[int]
        :param team_id: Team ID where the artifacts are stored. If not provided, will be taken from the current context.
        :type team_id: Optional[int]
        :param workspace_id: Workspace ID where the app will be deployed. If not provided, will be taken from the current context.
        :type workspace_id: Optional[int]
        :param agent_id: Agent ID. If not provided, will be found automatically.
        :type agent_id: Optional[int]
        :param kwargs: Additional parameters to start the task. See Api.task.start() for more details.
        :type kwargs: Dict[str, Any]
        :return: Task Info
        :rtype: Dict[str, Any]
        :raises ValueError: if validations fail.
        """
        from supervisely.nn.utils import ModelSource

        if not isinstance(artifacts_dir, str) or not artifacts_dir.strip():
            raise ValueError("artifacts_dir must be a non-empty string.")

        runtime = get_runtime(runtime)
        if team_id is None:
            team_id = self._find_team_by_path(artifacts_dir, team_id=team_id)
        logger.debug(
            f"Starting custom model deployment. Team: {team_id}, Artifacts Dir: '{artifacts_dir}'"
        )
        if agent_id is None:
            agent_id = self._find_agent()

        # Train V1 logic (if artifacts_dir does not start with '/experiments')
        if not artifacts_dir.startswith("/experiments"):
            logger.debug("Deploying model from Train V1 artifacts")
            module_id, serve_app_name, deploy_params = self._deploy_params_v1(
                team_id, artifacts_dir, checkpoint_name, device, runtime, with_module=True
            )
        else:  # Train V2 logic (when artifacts_dir starts with '/experiments')
            logger.debug("Deploying model from Train V2 artifacts")

            module_id, serve_app_name, deploy_params = self._deploy_params_v2(
                team_id, artifacts_dir, checkpoint_name, device, runtime, with_module=True
            )
        deploy_params["model_source"] = ModelSource.CUSTOM

        logger.info(
            f"{serve_app_name} app deployment started. Checkpoint: '{checkpoint_name}'. Deploy params: '{deploy_params}'"
        )
        try:
            task_info = self._run_serve_app(
                agent_id, module_id, workspace_id=workspace_id, **kwargs
            )
            self._load_model_from_api(task_info["id"], deploy_params)
        except Exception as e:
            raise RuntimeError(f"Failed to run '{serve_app_name}': {e}") from e
        return task_info

    def deploy_custom_model_from_experiment_info(
        self,
        agent_id: int,
        experiment_info: "ExperimentInfo",
        checkpoint_name: Optional[str] = None,
        device: Optional[str] = None,
        runtime: str = None,
        timeout: int = 100,
        **kwargs,
    ) -> Dict[str, Any]:
        """
        Deploy a custom model based on the training session.

        :param experiment_info: an :class:`ExperimentInfo` object.
        :type experiment_info: ExperimentInfo
        :param checkpoint_name: Checkpoint name (with file extension) to deploy, e.g. "best.pt".
            If not provided, the best checkpoint will be chosen.
        :type checkpoint_name: Optional[str]
        :param device: Device string. If not provided, will be chosen automatically.
        :type device: Optional[str]
        :param timeout: Timeout in seconds (default is 100). The maximum time to wait for the serving app to be ready.
        :type timeout: Optional[int]
        :param kwargs: Additional parameters to start the task. See Api.task.start() for more details.
        :type kwargs: Dict[str, Any]
        :return: Task Info
        :rtype: Dict[str, Any]
        :raises ValueError: if validations fail.
        """
        task_id = experiment_info.task_id
        train_task_info = self._api.task.get_info_by_id(task_id)
        runtime = get_runtime(runtime)

        logger.debug(f"Starting model deployment from experiment info. Task ID: '{task_id}'")
        train_module_id = train_task_info["meta"]["app"]["moduleId"]
        module = self.get_serving_app_by_train_app(module_id=train_module_id)
        serve_app_name = module["name"]
        module_id = module["id"]
        logger.debug(f"Serving app detected: '{serve_app_name}'. Module ID: '{module_id}'")

        if checkpoint_name is None:
            checkpoint_name = experiment_info.best_checkpoint

        # Task parameters
        experiment_name = experiment_info.experiment_name
        task_name = experiment_name + f" ({checkpoint_name})"
        if "task_name" not in kwargs:
            kwargs["task_name"] = task_name

        description = f"""Serve from experiment
                Experiment name:   {experiment_name}
                Evaluation report: {experiment_info.evaluation_report_link}
            """
        while len(description) > 255:
            description = description.rsplit("\n", 1)[0]
        if "description" not in kwargs:
            kwargs["description"] = description

        logger.info(f"{serve_app_name} app deployment started. Checkpoint: '{checkpoint_name}'.")
        try:
            task_info = self._run_serve_app(agent_id, module_id, timeout=timeout, **kwargs)
            self.load_custom_model_from_experiment_info(
                task_info["id"], experiment_info, checkpoint_name, device, runtime
            )
        except Exception as e:
            raise RuntimeError(f"Failed to run '{serve_app_name}': {e}") from e
        return task_info

    def start_serve_app(
        self, agent_id: int, app_name=None, module_id=None, **kwargs
    ) -> Dict[str, Any]:
        """
        Run a serving app. Either app_name or module_id must be provided.

        :param app_name: App name in Supervisely.
        :type app_name: Optional[str]
        :param module_id: Module ID in Supervisely.
        :type module_id: Optional[int]
        :param kwargs: Additional parameters to start the task. See Api.task.start() for more details.
        :type kwargs: Dict[str, Any]
        :return: Task Info
        :rtype: Dict[str, Any]
        """
        if app_name is None and module_id is None:
            raise ValueError("Either app_name or module_id must be provided.")
        if app_name is not None and module_id is not None:
            raise ValueError("Only one of app_name or module_id must be provided.")
        if module_id is None:
            module_id = self._api.app.find_module_id_by_app_name(app_name)
        self._run_serve_app(agent_id, module_id, **kwargs)

    def _run_serve_app(
        self, agent_id: int, module_id, workspace_id: int = None, timeout: int = 100, **kwargs
    ):
        _attempt_delay_sec = 1
        _attempts = timeout // _attempt_delay_sec

        # @TODO: Run app in team?
        if workspace_id is None:
            workspace_id = env.workspace_id()
        kwargs = get_valid_kwargs(
            kwargs=kwargs,
            func=self._api.task.start,
            exclude=["self", "module_id", "workspace_id", "agent_id"],
        )
        task_info = self._api.task.start(
            agent_id=agent_id,
            module_id=module_id,
            workspace_id=workspace_id,
            **kwargs,
        )
        ready = self._api.app.wait_until_ready_for_api_calls(
            task_info["id"], _attempts, _attempt_delay_sec
        )
        if not ready:
            raise TimeoutError(
                f"Task {task_info['id']} is not ready for API calls after {timeout} seconds."
            )
        return task_info

    def _load_model_from_api(self, task_id, deploy_params, model_name: Optional[str] = None):
        logger.info("Loading model")
        self._api.task.send_request(
            task_id,
            "deploy_from_api",
            data={"deploy_params": deploy_params, "model_name": model_name},
            raise_error=True,
        )
        logger.info("Model loaded successfully")

    def find_serving_app_by_framework(self, framework: str):
        modules = self._api.app.get_list_ecosystem_modules(
            categories=["serve", f"framework:{framework}"], categories_operation="and"
        )
        if len(modules) == 0:
            return None
        return modules[0]

    def find_serving_app_by_slug(self, slug: str) -> int:
        return self._api.app.get_ecosystem_module_id(slug)

    def get_serving_app_by_train_app(self, app_name: Optional[str] = None, module_id: int = None):
        if app_name is None and module_id is None:
            raise ValueError("Either app_name or module_id must be provided.")
        if app_name is not None:
            module_id = self._api.app.find_module_id_by_app_name(app_name)
        train_module_info = self._api.app.get_ecosystem_module_info(module_id)
        train_app_config = train_module_info.config
        categories = train_app_config["categories"]
        framework = None
        for category in categories:
            if category.lower().startswith("framework:"):
                framework = category.lstrip("framework:")
                break
        if framework is None:
            raise ValueError(
                "Unable to define serving app. Framework is not specified in the train app"
            )

        logger.debug(f"Detected framework: {framework}")
        module = self.find_serving_app_by_framework(framework)
        if module is None:
            raise ValueError(f"No serving apps found for framework {framework}")
        return module

    def get_deploy_info(self, task_id: int) -> Dict[str, Any]:
        """
        Get deploy info of a serving task.

        :param task_id: Task ID of the serving App.
        :type task_id: int
        :return: Deploy Info
        :rtype: Dict[str, Any]
        """
        return self._api.task.send_request(task_id, "get_deploy_info", data={}, raise_error=True)

    def _deploy_params_v1(
        self,
        team_id: int,
        artifacts_dir: str,
        checkpoint_name: str,
        device: str,
        runtime: str,
        with_module: bool = True,
    ) -> Tuple[int, Dict[str, Any]]:
        from supervisely.nn.artifacts import RITM, YOLOv5
        from supervisely.nn.artifacts.artifacts import BaseTrainArtifacts
        from supervisely.nn.utils import ModelSource

        framework_cls = self._get_framework_by_path(artifacts_dir)
        if not framework_cls:
            raise ValueError(f"Unsupported framework for artifacts_dir: '{artifacts_dir}'")

        framework: BaseTrainArtifacts = framework_cls(team_id)
        if framework_cls is RITM or framework_cls is YOLOv5:
            raise ValueError(
                f"{framework.framework_name} framework is not supported for deployment"
            )

        runtime = get_runtime(runtime)
        logger.debug(f"Detected framework: '{framework.framework_name}'")
        module_id = None
        serve_app_name = None
        if with_module:
            module_id = self._api.app.get_ecosystem_module_id(framework.serve_slug)
            serve_app_name = framework.serve_app_name
            logger.debug(f"Module ID fetched:' {module_id}'. App name: '{serve_app_name}'")

        train_info = framework.get_info_by_artifacts_dir(artifacts_dir.rstrip("/"))
        if not hasattr(train_info, "checkpoints") or not train_info.checkpoints:
            raise ValueError("No checkpoints found in train info.")

        checkpoint = None
        if checkpoint_name is not None:
            for cp in train_info.checkpoints:
                if cp.name == checkpoint_name:
                    checkpoint = cp
                    break
            if checkpoint is None:
                raise ValueError(f"Checkpoint '{checkpoint_name}' not found in train info.")
        else:
            logger.info("Checkpoint name not provided. Using the last checkpoint.")
            checkpoint = train_info.checkpoints[-1]

        checkpoint_name = checkpoint.name
        deploy_params = {
            "device": device,
            "model_source": ModelSource.CUSTOM,
            "task_type": train_info.task_type,
            "checkpoint_name": checkpoint_name,
            "checkpoint_url": checkpoint.path,
        }

        if getattr(train_info, "config_path", None) is not None:
            deploy_params["config_url"] = train_info.config_path

        if framework.require_runtime:
            deploy_params["runtime"] = runtime
        return module_id, serve_app_name, deploy_params

    def _deploy_params_v2(
        self,
        team_id: int,
        artifacts_dir: str,
        checkpoint_name: str,
        device: str,
        runtime: Optional[str] = None,
        with_module: bool = True,
    ):
        from supervisely.nn.experiments import get_experiment_info_by_artifacts_dir
        from supervisely.nn.utils import ModelSource

        experiment_info = get_experiment_info_by_artifacts_dir(self._api, team_id, artifacts_dir)
        if not experiment_info:
            raise ValueError(
                f"Failed to retrieve experiment info for artifacts_dir: '{artifacts_dir}'"
            )

        runtime = get_runtime(runtime)
        experiment_task_id = experiment_info.task_id
        experiment_task_info = self._api.task.get_info_by_id(experiment_task_id)
        if experiment_task_info is None:
            raise ValueError(f"Task with ID '{experiment_task_id}' not found")

        module_id = None
        serve_app_name = None
        if with_module:
            train_module_id = experiment_task_info["meta"]["app"]["moduleId"]
            module = self.get_serving_app_by_train_app(module_id=train_module_id)
            serve_app_name = module["name"]
            module_id = module["id"]
            logger.debug(f"Serving app detected: '{serve_app_name}'. Module ID: '{module_id}'")

        if len(experiment_info.checkpoints) == 0:
            raise ValueError(f"No checkpoints found in: '{artifacts_dir}'.")

        checkpoint = None
        if checkpoint_name is not None:
            if checkpoint_name.endswith(".pt") or checkpoint_name.endswith(".pth"):
                for checkpoint_path in experiment_info.checkpoints:
                    if get_file_name_with_ext(checkpoint_path) == checkpoint_name:
                        checkpoint = get_file_name_with_ext(checkpoint_path)
                        break
            elif checkpoint_name.endswith(".onnx"):
                checkpoint_path = experiment_info.export.get("ONNXRuntime")
                if checkpoint_path is None:
                    raise ValueError(f"ONNXRuntime export not found in: '{artifacts_dir}'.")
            elif checkpoint_name.endswith(".engine"):
                checkpoint_path = experiment_info.export.get("TensorRT")
                if checkpoint_path is None:
                    raise ValueError(f"TensorRT export not found in: '{artifacts_dir}'.")
            else:
                raise ValueError(
                    f"Unknown checkpoint format: '{checkpoint_name}'. Expected formats: '.pt', '.pth', '.onnx' or '.engine'"
                )

            checkpoint = get_file_name_with_ext(checkpoint_path)
            if checkpoint is None:
                raise ValueError(f"Provided checkpoint '{checkpoint_name}' not found")
        else:
            logger.info("Checkpoint name not provided. Using the best checkpoint.")
            checkpoint = experiment_info.best_checkpoint

        model_info_dict = asdict(experiment_info)
        model_info_dict["artifacts_dir"] = artifacts_dir
        checkpoint_name = checkpoint
        checkpoints_dir = self._get_checkpoints_dir(checkpoint_name)
        checkpoint_path = f"/{artifacts_dir.strip('/')}/{checkpoints_dir}/{checkpoint_name}"
        if runtime is None:
            runtime = self._set_auto_runtime_by_checkpoint(checkpoint_path)

        deploy_params = {
            "device": device,
            "model_source": ModelSource.CUSTOM,
            "model_files": {"checkpoint": checkpoint_path},
            "model_info": model_info_dict,
            "runtime": runtime,
        }

        config = experiment_info.model_files.get("config")
        if config is not None:
            deploy_params["model_files"]["config"] = f"{experiment_info.artifacts_dir}{config}"
            logger.debug(f"Config file added: {experiment_info.artifacts_dir}{config}")
        return module_id, serve_app_name, deploy_params

    def _set_auto_runtime_by_checkpoint(self, checkpoint_path: str) -> str:
        if checkpoint_path.endswith(".pt") or checkpoint_path.endswith(".pth"):
            return RuntimeType.PYTORCH
        elif checkpoint_path.endswith(".onnx"):
            return RuntimeType.ONNXRUNTIME
        elif checkpoint_path.endswith(".engine"):
            return RuntimeType.TENSORRT
        else:
            raise ValueError(f"Unknown checkpoint format: '{checkpoint_path}'")

    def wait(self, model_id, target: Literal["started", "deployed"] = "started", timeout=5 * 60):
        t = time.monotonic()
        method = "is_alive" if target == "started" else "is_ready"
        while time.monotonic() - t < timeout:
            self._api.task.send_request(model_id, "is_ready", {})
            time.sleep(1)

    def _get_artifacts_dir_and_checkpoint_name(self, model: str) -> Tuple[str, str]:
        if not model.startswith("/"):
            raise ValueError(f"Path must start with '/'")

        if model.startswith("/experiments"):
            if model.endswith(".pt") or model.endswith(".pth"):
                try:
                    artifacts_dir, checkpoint_name = model.split("/checkpoints/")
                    return artifacts_dir, checkpoint_name
                except:
                    raise ValueError(
                        "Bad format of checkpoint path. Expected format: '/artifacts_dir/checkpoints/checkpoint_name'"
                    )
            elif model.endswith(".onnx") or model.endswith(".engine"):
                try:
                    artifacts_dir, checkpoint_name = model.split("/export/")
                    return artifacts_dir, checkpoint_name
                except:
                    raise ValueError(
                        "Bad format of checkpoint path. Expected format: '/artifacts_dir/export/checkpoint_name'"
                    )
            else:
                raise ValueError(f"Unknown model format: '{get_file_name_with_ext(model)}'")

        framework_cls = self._get_framework_by_path(model)
        if framework_cls is None:
            raise ValueError(f"Unknown path: '{model}'")

        team_id = env.team_id()
        framework = framework_cls(team_id)
        checkpoint_name = get_file_name_with_ext(model)
        checkpoints_dir = model.replace(checkpoint_name, "")
        if framework.weights_folder is not None:
            artifacts_dir = checkpoints_dir.replace(framework.weights_folder, "")
        else:
            artifacts_dir = checkpoints_dir
        return artifacts_dir, checkpoint_name

    def _get_checkpoints_dir(self, checkpoint_name: str) -> str:
        if checkpoint_name.endswith(".pt") or checkpoint_name.endswith(".pth"):
            return "checkpoints"
        elif checkpoint_name.endswith(".onnx") or checkpoint_name.endswith(".engine"):
            return "export"
        else:
            raise ValueError(f"Unknown checkpoint format: '{checkpoint_name}'")

    def _get_framework_by_path(self, path: str):
        from supervisely.nn.artifacts import (
            RITM,
            RTDETR,
            Detectron2,
            MMClassification,
            MMDetection,
            MMDetection3,
            MMSegmentation,
            UNet,
            YOLOv5,
            YOLOv5v2,
            YOLOv8,
        )
        from supervisely.nn.artifacts.artifacts import BaseTrainArtifacts
        from supervisely.nn.utils import ModelSource

        path_obj = Path(path)
        if len(path_obj.parts) < 2:
            raise ValueError(f"Incorrect checkpoint path: '{path}'")
        parent = path_obj.parts[1]
        frameworks = {
            "/detectron2": Detectron2,
            "/mmclassification": MMClassification,
            "/mmdetection": MMDetection,
            "/mmdetection-3": MMDetection3,
            "/mmsegmentation": MMSegmentation,
            "/RITM_training": RITM,
            "/RT-DETR": RTDETR,
            "/unet": UNet,
            "/yolov5_train": YOLOv5,
            "/yolov5_2.0_train": YOLOv5v2,
            "/yolov8_train": YOLOv8,
        }
        if f"/{parent}" in frameworks:
            return frameworks[f"/{parent}"]



# Pointcloud
from supervisely.convert.pointcloud.sly.sly_pointcloud_converter import SLYPointcloudConverter
from supervisely.convert.pointcloud.las.las_converter import LasConverter
from supervisely.convert.pointcloud.ply.ply_converter import PlyConverter
from supervisely.convert.pointcloud.bag.bag_converter import BagConverter
from supervisely.convert.pointcloud.lyft.lyft_converter import LyftConverter
from supervisely.convert.pointcloud.nuscenes_conv.nuscenes_converter import NuscenesConverter
from supervisely.convert.pointcloud.kitti_3d.kitti_3d_converter import KITTI3DConverter



from supervisely.api.project_api import ProjectInfo
from supervisely.app.widgets import (
    Button,
    Card,
    Checkbox,
    Container,
    ProjectThumbnail,
    Text,
)
from supervisely.project.download import is_cached


class InputSelector:
    title = "Input project"
    description = "Selected project from which items and annotations will be downloaded"
    lock_message = None

    def __init__(self, project_info: ProjectInfo, app_options: dict = {}):
        # Init widgets
        self.project_thumbnail = None
        self.use_cache_text = None
        self.use_cache_checkbox = None
        self.validator_text = None
        self.button = None
        self.container = None
        self.card = None
        # -------------------------------- #

        self.display_widgets = []
        self.app_options = app_options

        self.project_id = project_info.id
        self.project_info = project_info

        # GUI Components
        self.project_thumbnail = ProjectThumbnail(self.project_info)

        if is_cached(self.project_id):
            _text = "Use cached data stored on the agent to optimize project download"
        else:
            _text = "Cache data on the agent to optimize project download for future trainings"
        self.use_cache_text = Text(_text)
        self.use_cache_checkbox = Checkbox(self.use_cache_text, checked=True)

        self.validator_text = Text("")
        self.validator_text.hide()
        self.button = Button("Select")
        self.display_widgets.extend(
            [
                self.project_thumbnail,
                self.use_cache_checkbox,
                self.validator_text,
                self.button,
            ]
        )
        # -------------------------------- #

        self.container = Container(self.display_widgets)
        self.card = Card(
            title=self.title,
            description=self.description,
            content=self.container,
            collapsable=self.app_options.get("collapsable", False),
        )

    @property
    def widgets_to_disable(self) -> list:
        return [self.use_cache_checkbox]

    def get_project_id(self) -> int:
        return self.project_id

    def set_cache(self, value: bool) -> None:
        if value:
            self.use_cache_checkbox.check()
        else:
            self.use_cache_checkbox.uncheck()

    def get_cache_value(self) -> bool:
        return self.use_cache_checkbox.is_checked()

    def validate_step(self) -> bool:
        return True



from typing import Optional, Union, List

from supervisely import ObjClass, ObjClassCollection
from supervisely.app import StateJson
from supervisely.app.widgets import Widget


class ClassesMappingPreview(Widget):
    def __init__(
        self,
        classes: Optional[Union[List[ObjClass], ObjClassCollection]] = [],
        mapping: Optional[dict] = {},
        max_height: str = "128px",
        widget_id: Optional[str] = None,
    ):
        self._classes = classes
        self._mapping = mapping
        self._max_height = max_height
        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self):
        return {
            "max_height": self._max_height,
        }

    def get_json_state(self):
        return {
            "mapping": [
                {"class": cls.to_json(), "value": self._mapping.get(cls.name, "")}
                for cls in self._classes
            ]
        }

    def set(self, classes: Union[List[ObjClass], ObjClassCollection], mapping: dict):
        self._mapping = {
            k: v["value"]
            for k, v in mapping.items()
            if k in [obj_class.name for obj_class in classes]
        }
        self._classes = [obj_class for obj_class in classes if obj_class.name in self._mapping]
        StateJson()[self.widget_id] = self.get_json_state()
        StateJson().send_changes()

    def set_mapping(self, mapping: dict):
        if self._classes is None:
            raise ValueError("Classes are not set")
        self._mapping = {
            k: v["value"]
            for k, v in mapping.items()
            if k in [obj_class.name for obj_class in self._classes]
        }
        StateJson()[self.widget_id] = self.get_json_state()
        StateJson().send_changes()

    def get_classes(self):
        return self._classes

    def get_mapping(self):
        return self._mapping



from supervisely.app.widgets import Widget
from supervisely.app import DataJson
from supervisely.api.api import Api
from supervisely.api.video.video_api import VideoInfo
from supervisely.project.project import Project
from supervisely.video.video import get_labeling_tool_url, get_labeling_tool_link


class VideoThumbnail(Widget):
    def __init__(self, info: VideoInfo = None, widget_id: str = None):
        self._info: VideoInfo = None
        self._id: int = None
        self._name: str = None
        self._description: str = None
        self._url: str = None
        self._open_link: str = None
        self._preview_url: str = None
        self._set_info(info)

        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self):
        return {
            "id": self._id,
            "name": self._name,
            "description": self._description,
            "url": self._url,
            "link": self._open_link,
            "image_preview_url": self._preview_url,
        }

    def _set_info(self, info: VideoInfo = None):
        if info is None:
            return
        self._info = info
        self._id = info.id
        self._name = info.name
        self._description = (
            f"Video length: {info.duration_hms} / {info.frames_count_compact} frames"
        )
        self._url = get_labeling_tool_url(info.dataset_id, info.id)
        self._open_link = get_labeling_tool_link(self._url, "open video")
        self._preview_url = info.image_preview_url

    def set_video(self, info: VideoInfo):
        self._set_info(info)
        self.update_data()
        DataJson().send_changes()

    def set_video_id(self, id: int):
        api = Api()
        info = api.video.get_info_by_id(id, raise_error=True)
        self.set_video(info)

    def get_json_state(self):
        return None
